/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlObjectDetectionAnalyzer.h
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
**              based on applications by Stephen Gould <sgould@stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
** DESCRIPTION:
**  Provides helper functions for analyzing the detections of the
**  sliding window classifier.
**
*****************************************************************************/

#pragma once

#include <cassert>
#include <iostream>
#include <vector>
#include <map>

#include "cv.h"
#include "cxcore.h"
#include "ml.h"

#include "svlBase.h"
#include "svlObjectList.h"
#include "svlML.h"

using namespace std;

// svlObjectDetectionAnalyzer

static const string NO_NAME = "";

class svlObjectDetectionAnalyzer {
 public:
  static float DEFAULT_AREA_RATIO;
  static bool USE_NONMAX;
  static bool INCLUDE_ALL_FRAMES;

 private:
  // if just interested in one object
  string _objectName;  
  svlClassifierResponseList _objectScores;
  svlPRcurve _PRcurve;

 protected:

  svlObject2dSequence _gtObjectList;
  svlObject2dSequence _objectList;
  bool _useNonmax;

 public:
  svlObjectDetectionAnalyzer(string name = NO_NAME) : _objectName(name) {}
    
    // to suppress warnings
    virtual ~svlObjectDetectionAnalyzer() {}

  // loads the gt file; discards all that don't match 'object_name'
  // if it was set
  virtual bool loadGt(const char *filename);

  // careful with this, make sure names match
  void copyGt(svlObjectDetectionAnalyzer &d) {
    _gtObjectList = d._gtObjectList;
  }

  void setNonmax(bool useNonmax) { _useNonmax = useNonmax; }

  // loads the detection file, discards all that don't match 'object_name'
  // if it was set, and filters non-maximal objects
  virtual bool loadDetections(const char *filename);
  virtual void loadDetections(const svlObject2dSequence & s);

  svlObject2dSequence getDetections() const { return _objectList; }

  // based on the groundtruth and detections that are currently loaded
  virtual void accumulateScores(bool ignoreMissingFrames = !INCLUDE_ALL_FRAMES,
				float areaRatio = DEFAULT_AREA_RATIO);

  // will not actually store the detections, but will continuously
  // accumulate objectScores
  bool accumulateScoresForFrame(string frameName,
				svlObject2dFrame &objects,
				float areaRatio = DEFAULT_AREA_RATIO);

  // based on the accumulated scores, prints a summary of results
  virtual void printResults(float threshold) const;

  // given the current accumulated scores, creates a PR curve
  virtual void createPRcurves();

  // same as above, except a rough approximation
  virtual void createDiscretizedPRcurves(float dthreshold = DEFAULT_THRESHOLD_DELTA);

  // prints all the PR curves currently stored
  virtual void writePRcurves(const char *filestem) const;

  // loads the prCurve from file
  virtual bool loadPRcurve(const char *filename);

  // gets the highest F-score points on the PR curve (must be already created)
  float computeHighestFscore() const;

  // gets the area of the currently stored PR curve
  float computePRarea() const;

  virtual void beginAccumulatingScores();

  // outputs after loading detections
  void outputDetections(const char *filename, float threshold = 0);
};


typedef map<string, svlClassifierResponseList> TResponses;
typedef map<string, svlPRcurve> TCurves;

class svlMultObjectsDetectionAnalyzer : public svlObjectDetectionAnalyzer {
 private:
  TResponses _allObjectScores;
  TCurves _allPRcurves;

 public:
  // to suppress warnings
  ~svlMultObjectsDetectionAnalyzer() {}

  // defining virtual functions from above
  bool loadGt(const char *filename);
  bool loadDetections(const char *filename);
  void accumulateScores(bool ignoreMissingFrames, float areaRatio = DEFAULT_AREA_RATIO);
  void printResults(float threshold) const;
  void createPRcurves();
  void createDiscretizedPRcurves(float dthreshold = DEFAULT_THRESHOLD_DELTA);
  void writePRcurves(const char *filestem) const;

  // loads a PR curve with a specific name
  bool loadPRcurve(const char *filename); // will return false
  bool loadPRcurve(const char *filename, string name);

  // gets the highest F-score points on one of the PR curves
  // (must be already created); name doesn't need to be specified
  // iff there's only one curve currently stored
  float computeHighestFscore(string name = NO_NAME);

  // gets the area of one of the currently stored PR curves
  float computePRarea(string name = NO_NAME);

  void beginAccumulatingScores();
};
